@@ -27,12 +27,13 @@
 import re
 import threading
 import time
+import uuid
 try:
     from urlparse import urlparse
 except ImportError:
     from urllib.parse import urlparse
 
-from msrest.exceptions import DeserializationError
+from msrest.exceptions import DeserializationError, ClientException
 from msrestazure.azure_exceptions import CloudError
 
 
@@ -59,6 +60,33 @@ def succeeded(status):
     return str(status).lower() in SUCCEEDED
 
 
+def _validate(url):
+    """Validate a url.
+
+    :param str url: Polling URL extracted from response header.
+    :raises: ValueError if URL has no scheme or host.
+    """
+    if url is None:
+        return
+    parsed = urlparse(url)
+    if not parsed.scheme or not parsed.netloc:
+        raise ValueError("Invalid URL header")
+
+def _get_header_url(response, header_name):
+    """Get a URL from a header requests.
+
+    :param requests.Response response: REST call response.
+    :param str header_name: Header name.
+    :returns: URL if not None AND valid, None otherwise
+    """
+    url = response.headers.get(header_name)
+    try:
+        _validate(url)
+    except ValueError:
+        return None
+    else:
+        return url
+
 class BadStatus(Exception):
     pass
 
@@ -71,10 +99,6 @@ class OperationFailed(Exception):
     pass
 
 
-class OperationFinished(Exception):
-    pass
-
-
 class SimpleResource:
     """An implementation of Python 3 SimpleNamespace.
     Used to deserialize resource objects from response bodies where
@@ -107,30 +131,18 @@ class LongRunningOperation(object):
         self.get_outputs = outputs
         self.async_url = None
         self.location_url = None
+        self.initial_status_code = None
 
-    def _validate(self, url):
-        """Validate header url
-
-        :param str url: Polling URL extracted from response header.
-        :returns: URL if valid.
-        :raises: ValueError if URL has not scheme or host.
-        """
-        if url is None:
-            return None
-        parsed = urlparse(url)
-        if not parsed.scheme or not parsed.netloc:
-            raise ValueError("Invalid URL header")
-        return url
-
-    def _check_status(self, response):
+    def _raise_if_bad_http_status_and_method(self, response):
         """Check response status code is valid for a Put or Patch
-        reqest. Must be 200, 202, or 204.
+        request. Must be 200, 201, 202, or 204.
 
         :raises: BadStatus if invalid status.
         """
         code = response.status_code
-        if code in [200, 202] or (code == 201 and self.method == 'PUT') or \
-                (code == 204 and self.method in ['DELETE', 'POST']):
+        if code in {200, 202} or \
+           (code == 201 and self.method in {'PUT', 'PATCH'}) or \
+           (code == 204 and self.method in {'DELETE', 'POST'}):
             return
         raise BadStatus(
             "Invalid return status for {!r} operation".format(self.method))
@@ -154,23 +166,30 @@ class LongRunningOperation(object):
     def _deserialize(self, response):
         """Attempt to deserialize resource from response.
 
-        :raises: OperationFailed if deserialized resource has status of
-         failed or cancelled.
-        :raises: OperationFinished if deserialised resource has status
-         succeeded.
+        :param requests.Response response: latest REST call response.
         """
-        self.resource = self.get_outputs(response)
-        if self.method == 'PUT':
-            resource_status = self._get_resource_status()
-            if failed(resource_status):
-                self.status = resource_status
-                raise OperationFailed("Operation failed or cancelled")
-            elif succeeded(resource_status):
-                raise OperationFinished("Operation succeeded")
-            elif resource_status:
-                self.status = resource_status
+        # Hacking response with initial status_code
+        previous_status = response.status_code
+        response.status_code = self.initial_status_code
+        resource = self.get_outputs(response)
+        response.status_code = previous_status
 
-    def _get_body_status(self, response):
+        # Hack for Storage or SQL, to workaround the bug in the Python generator
+        if resource is None:
+            previous_status = response.status_code
+            for status_code_to_test in [200, 201]:
+                try:
+                    response.status_code = status_code_to_test
+                    resource = self.get_outputs(response)
+                except ClientException:
+                    pass
+                else:
+                    return resource
+                finally:
+                    response.status_code = previous_status
+        return resource
+
+    def _get_async_status(self, response):
         """Attempt to find status info in response body.
 
         :param requests.Response response: latest REST call response.
@@ -182,122 +201,61 @@ class LongRunningOperation(object):
         body = response.json()
         return body.get('status')
 
-    def _get_resource_status(self):
+    def _get_provisioning_state(self, response):
         """
         Attempt to get provisioning state from resource.
+        :param requests.Response response: latest REST call response.
         :returns: Status if found, else 'None'.
         """
-        try:
-            return self.resource.provisioning_state
-        except AttributeError:
-            pass
-        try:
-            return self.resource.properties.provisioning_state
-        except AttributeError:
+        if self._is_empty(response):
             return None
-
-    def _object_from_response(self, response):
-        """If deserialization fails, attempt to create object from
-        response body regardless.
-        Required functionality for Azure LRO's....
-
-        :param requests.Response response: latest REST call response.
-        """
         body = response.json()
-        body = {self._convert.sub(r'\1_\2', k).lower(): v
-                for k, v in body.items()}
-        properties = body.get('properties')
-        if properties:
-            properties = {self._convert.sub(r'\1_\2', k).lower(): v
-                          for k, v in properties.items()}
-            del body['properties']
-            body.update(properties)
-            self.resource = SimpleResource(**body)
-        else:
-            self.resource = SimpleResource(**body)
+        return body.get("properties", {}).get("provisioningState")
 
-    def _process_status(self, response):
-        """Process response based on specific status code.
+    def should_do_final_get(self):
+        """Check whether the polling should end doing a final GET.
 
         :param requests.Response response: latest REST call response.
-        """
-        process = getattr(self, '_status_' + str(response.status_code))
-        process(response)
-
-    def _status_200(self, response):
-        """Process response with status code 200.
-
-        :param requests.Response response: latest REST call response.
-        """
-        status = self._get_body_status(response)
-        self.status = status if status else 'Succeeded'
-        if not status:
-            try:
-                # Even if this fails, status '200' should be successful.
-                self._deserialize(response)
-            except CloudError:
-                if self.method in ['PUT', 'PATCH']:
-                    self._object_from_response(response)
-
-    def _status_201(self, response):
-        """Process response with status code 201.
-
-        :param requests.Response response: latest REST call response.
-        :raises: BadResponse if response deserializes to CloudError.
-        """
-        try:
-            self._deserialize(response)
-            if not self.status:
-                self.status = 'Succeeded'
-        except CloudError as err:
-            raise BadResponse(str(err))
-
-    def _status_202(self, response):
-        """Process response with status code 202.
-        Just sets status to 'InProgress'.
-
-        :param requests.Response response: latest REST call response.
-        """
-        self.status = 'InProgress'
-
-    def _status_204(self, response):
-        """Process response with status code 204.
-        Interpretted as successful with no payload.
-
-        :param requests.Response response: latest REST call response.
-        """
-        self.status = 'Succeeded'
-        self.resource = None
-
-    def is_done(self):
-        """Check whether the operation can be considered complete.
-        This is based on whether the data in the resource matches the current
-        status. If there is not resource, we assume it's complete.
-
         :rtype: bool
         """
-        if (self.async_url or not self.resource) and \
-                self.method in ['PUT', 'PATCH']:
-            return False
-        resource_state = self._get_resource_status()
-        try:
-            return self.status.lower() == resource_state.lower()
-        except AttributeError:
-            return True
+        return (self.async_url or not self.resource) and \
+                self.method in {'PUT', 'PATCH'}
 
-    def get_initial_status(self, response):
+    def set_initial_status(self, response):
         """Process first response after initiating long running
-        operation.
+        operation and set self.status attribute.
 
         :param requests.Response response: initial REST call response.
         """
-        self._deserialize(response)
-        self._check_status(response)
-        if response.status_code in [200, 202, 204]:
-            self._process_status(response)
-        status = self._get_body_status(response)
-        if status:
-            self.status = status
+        self._raise_if_bad_http_status_and_method(response)
+
+        if self._is_empty(response):
+            self.resource = None
+        else:
+            try:
+                self.resource = self.get_outputs(response)
+            except DeserializationError:
+                self.resource = None
+
+        self.set_async_url_if_present(response)
+
+        if response.status_code in {200, 201, 202, 204}:
+            self.initial_status_code = response.status_code
+            if self.async_url or self.location_url or response.status_code == 202:
+                self.status = 'InProgress'
+            elif response.status_code == 201:
+                status = self._get_provisioning_state(response)
+                self.status = status or 'InProgress'
+            elif response.status_code == 200:
+                status = self._get_provisioning_state(response)
+                self.status = status or 'Succeeded'
+            elif response.status_code == 204:
+                self.status = 'Succeeded'
+                self.resource = None
+            else:
+                raise OperationFailed("Invalid status found")
+            return
+        raise OperationFailed("Operation failed or cancelled")
 
     def get_status_from_location(self, response):
         """Process the latest status update retrieved from a 'location'
@@ -306,8 +264,16 @@ class LongRunningOperation(object):
         :param requests.Response response: latest REST call response.
         :raises: BadResponse if response has no body and not status 202.
         """
-        self._check_status(response)
-        self._process_status(response)
+        self._raise_if_bad_http_status_and_method(response)
+        code = response.status_code
+        if code == 202:
+            self.status = "InProgress"
+        else:
+            self.status = 'Succeeded'
+            if self._is_empty(response):
+                self.resource = None
+            else:
+                self.resource = self._deserialize(response)
 
     def get_status_from_resource(self, response):
         """Process the latest status update retrieved from the same URL as
@@ -316,11 +282,15 @@ class LongRunningOperation(object):
         :param requests.Response response: latest REST call response.
         :raises: BadResponse if status not 200 or 204.
         """
-        self._check_status(response)
-        if self._is_empty(response) and self.method in ['PUT', 'PATCH']:
-            raise BadResponse('The response from long running '
-                              'operation does not contain a body.')
-        self._process_status(response)
+        self._raise_if_bad_http_status_and_method(response)
+        if self._is_empty(response):
+            raise BadResponse('The response from long running operation '
+                              'does not contain a body.')
+
+        status = self._get_provisioning_state(response)
+        self.status = status or 'Succeeded'
+
+        self.resource = self._deserialize(response)
 
     def get_status_from_async(self, response):
         """Process the latest status update retrieved from a
@@ -330,42 +300,28 @@ class LongRunningOperation(object):
         :raises: BadResponse if response has no body, or body does not
          contain status.
         """
-        self._check_status(response)
+        self._raise_if_bad_http_status_and_method(response)
         if self._is_empty(response):
             raise BadResponse('The response from long running operation '
                               'does not contain a body.')
 
-        self.status = self._get_body_status(response)
+        self.status = self._get_async_status(response)
         if not self.status:
             raise BadResponse("No status found in body")
-        if self.method in ["POST", "DELETE"]:
-            try:
-                self._deserialize(response)
-            except CloudError:
-                pass  # Not all 'accept' statuses will deserialize.
 
-    def get_retry(self, response, *args):
-        """Retrieve the URL that will be polled for status. First looks for
-        'azure-asyncoperation' header, if not found or invalid, check for
-        'location' header.
-
-        :param requests.Response response: latest REST call response.
-        """
         try:
-            self.async_url = self._validate(
-                response.headers.get('azure-asyncoperation'))
+            self.resource = self.get_outputs(response)
+        except Exception:
+            self.resource = None
 
-            # Return if we have a url, in case location header raises error.
-            if self.async_url:
-                return
-        except ValueError:
-            pass  # We can ignore as location header may still be valid.
-        self.location_url = self._validate(response.headers.get('location'))
-        if not self.location_url and not self.async_url:
-            code = response.status_code
-            if code == 202 and self.method == 'POST':
-                raise BadResponse(
-                    'Location header is missing from long running operation.')
+    def set_async_url_if_present(self, response):
+        async_url = _get_header_url(response, 'azure-asyncoperation')
+        if async_url:
+            self.async_url = async_url
+        
+        location_url = _get_header_url(response, 'location')
+        if location_url:
+            self.location_url = location_url
 
 
 class AzureOperationPoller(object):
@@ -379,36 +335,45 @@ class AzureOperationPoller(object):
         of the operation.
     :param int timeout: Time in seconds to wait between status calls,
         default is 30.
-    :param callable func: Callback function that takes at least one
-        argument, a completed LongRunningOperation (optional).
     """
 
     def __init__(self, send_cmd, output_cmd, update_cmd, timeout=30):
         self._timeout = timeout
-        self._response = None
-        self._operation = None
-        self._exception = None
         self._callbacks = []
-        self._done = threading.Event()
-        self._thread = threading.Thread(
-            target=self._start, args=(send_cmd, update_cmd, output_cmd))
-        self._thread.daemon = True
-        self._thread.start()
 
-    def _start(self, send_cmd, update_cmd, output_cmd):
-        """Start the long running operation.
-        On completetion, runs any callbacks.
-
-        :param callable send_cmd: The API request to initiate the operation.
-        :param callable update_cmd: The API reuqest to check the status of
-         the operation.
-        :param callable output_cmd: The function to deserialize the resource
-         of the operation.
-        """
         try:
             self._response = send_cmd()
             self._operation = LongRunningOperation(self._response, output_cmd)
-            self._operation.get_initial_status(self._response)
+            self._operation.set_initial_status(self._response)
+        except BadStatus:
+            self._operation.status = 'Failed'
+            raise CloudError(self._response)
+        except BadResponse as err:
+            self._operation.status = 'Failed'
+            raise CloudError(self._response, str(err))
+        except OperationFailed:
+            raise CloudError(self._response)
+
+        self._thread = None
+        self._done = None
+        self._exception = None
+        if not finished(self.status()):
+            self._done = threading.Event()
+            self._thread = threading.Thread(
+                target=self._start,
+                name="AzureOperationPoller({})".format(uuid.uuid4()),
+                args=(update_cmd,))
+            self._thread.daemon = True
+            self._thread.start()
+
+    def _start(self, update_cmd):
+        """Start the long running operation.
+        On completetion, runs any callbacks.
+
+        :param callable update_cmd: The API reuqest to check the status of
+         the operation.
+        """
+        try:
             self._poll(update_cmd)
 
         except BadStatus:
@@ -422,9 +387,6 @@ class AzureOperationPoller(object):
         except OperationFailed:
             self._exception = CloudError(self._response)
 
-        except OperationFinished:
-            pass
-
         except Exception as err:
             self._exception = err
 
@@ -467,41 +429,52 @@ class AzureOperationPoller(object):
 
         :param callable update_cmd: The function to call to retrieve the
          latest status of the long running operation.
-        :raises: OperationFinished if operation status 'Succeeded'.
         :raises: OperationFailed if operation status 'Failed' or 'Cancelled'.
         :raises: BadStatus if response status invalid.
         :raises: BadResponse if response invalid.
         """
         initial_url = self._response.request.url
 
-        while not finished(self._operation.status):
+        while not finished(self.status()):
             self._delay()
-            url = self._response.request.url
             headers = self._polling_cookie()
-            self._operation.get_retry(self._response, initial_url)
 
             if self._operation.async_url:
                 self._response = update_cmd(
                     self._operation.async_url, headers)
+                self._operation.set_async_url_if_present(self._response)
                 self._operation.get_status_from_async(
                     self._response)
             elif self._operation.location_url:
                 self._response = update_cmd(
                     self._operation.location_url, headers)
+                self._operation.set_async_url_if_present(self._response)
                 self._operation.get_status_from_location(
                     self._response)
-            else:
-                self._response = update_cmd(url, headers)
+            elif self._operation.method == "PUT":
+                self._response = update_cmd(initial_url, headers)
+                self._operation.set_async_url_if_present(self._response)
                 self._operation.get_status_from_resource(
                     self._response)
+            else:
+                raise BadResponse(
+                    'Location header is missing from long running operation.')
 
         if failed(self._operation.status):
             raise OperationFailed("Operation failed or cancelled")
-        elif not self._operation.is_done():
+        elif self._operation.should_do_final_get():
             self._response = update_cmd(initial_url)
             self._operation.get_status_from_resource(
                 self._response)
 
+    def status(self):
+        """Returns the current status string.
+
+        :returns: The current status string
+        :rtype: str
+        """
+        return self._operation.status
+
     def result(self, timeout=None):
         """Return the result of the long running operation, or
         the result available after the specified timeout.
@@ -521,6 +494,8 @@ class AzureOperationPoller(object):
          operation to complete.
         :raises CloudError: Server problem with the query.
         """
+        if self._thread is None:
+            return
         self._thread.join(timeout=timeout)
         try:
             raise self._exception
@@ -532,7 +507,7 @@ class AzureOperationPoller(object):
 
         :returns: 'True' if the process has completed, else 'False'.
         """
-        return not self._thread.isAlive()
+        return self._thread is None or not self._thread.isAlive()
 
     def add_done_callback(self, func):
         """Add callback function to be run once the long running operation
@@ -543,7 +518,7 @@ class AzureOperationPoller(object):
         :raises: ValueError if the long running operation has already
          completed.
         """
-        if self._done.is_set():
+        if self._done is None or self._done.is_set():
             raise ValueError("Process is complete.")
         self._callbacks.append(func)
 
@@ -554,6 +529,6 @@ class AzureOperationPoller(object):
         :raises: ValueError if the long running operation has already
          completed.
         """
-        if self._done.is_set():
+        if self._done is None or self._done.is_set():
             raise ValueError("Process is complete.")
         self._callbacks = [c for c in self._callbacks if c != func]
